(function(global, factory) {

    /* AMD */ if (typeof define === 'function' && define["amd"])
        define(["protobuf", "pbrpc"], factory);
    /* CommonJS */ else if (typeof require === "function" && typeof module === "object" && module && module["exports"])
        module["exports"] = factory(require("protobuf"), require("pbrpc"), true);
    /* Global */ else
        (global["deepnet"] = global["deepnet"] || {})["DevicePass"] = factory(global["dcodeIO"]["ProtoBuf"], global["deepnet"]["PBRPC"]);

})(this, function(ProtoBuf, PBRPC) {

    "use strict";

    var DevicePass = {};

    if (!window.console) {
        window.console = {
            log: function () {
            }
        }
    }

    var start_port = 14800;
    var end_port = 14810;

    DevicePass.port = start_port;
    DevicePass.fqdn = "local-device-manager.dualshield.local";

    function scan_port(onOK, onFail) {
        ProtoBuf.Util.fetch("https://" + DevicePass.fqdn + ":" + DevicePass.port + "/ping_dm", function (content) {
            if (content != null){
                var parts = content.split(" ");
                if(parts.length>=2){
                    var id = parts[0];
                    var version = parts[1];
                    if(id=="DualShield_Device_Manager") {
                        onOK(version);
                        return;
                    }
                }
            }

            if(DevicePass.port >= end_port)
                onFail("DualShield Device Manager is not installed or started.");
            else {
                DevicePass.port = DevicePass.port + 1;
                scan_port(onOK, onFail);
            }
        });
    }

    DevicePass.initialize = function(resourceBase, onOK, onError){

        DevicePass.port = start_port;

        scan_port(function(v){

            DevicePass.version = v;
            console.log("DualShield Device Manager " + v + " on port " + DevicePass.port );

            PBRPC.initialize(resourceBase, function() {

                PBRPC.loadFile('device_pass', function (err, builder) {
                    if(err) {
                        console.log("err=" + err);
                        onError(err);
                        return;
                    }

                    DevicePass.DeviceInfo = builder.build('device_manager.proto.device_pass.DeviceInfo');
                    DevicePass.QueryCriteria = builder.build('device_manager.proto.device_pass.QueryCriteria');
                    DevicePass.Rejection = builder.build('device_manager.proto.device_pass.Rejection');

                    PBRPC.connect('wss://'  + DevicePass.fqdn + ':' + DevicePass.port + '/ws',
                        function(){
                            onOK(v);
                        },
                        onError);
                });
            });
        },
        function(e){
            onError(e);
        });

    };



    var _userOnReady;
    var _userOnDenied;
    var _userApplyTo;

    function _OnReady(response){
        if(_userOnReady)
            _userOnReady.call(_userApplyTo, DevicePass.DeviceInfo.decode(response));
    }

    function _OnDenied(response){
        if(_userOnDenied)
            _userOnDenied.call(_userApplyTo) ;
    }

    PBRPC.registerEventListener("on_device_print_ready", _OnReady);

    PBRPC.registerEventListener("on_device_access_denied", _OnDenied);

    DevicePass.MODE_VERIFY = 0;
    DevicePass.MODE_REGISTER = 1;
    DevicePass.MODE_REGISTER_WITH_INFO = 2;

    DevicePass.query = function(session, mode, lockToDomain, applyTo, onready, ondenied, onerror, onpopup) {

        _userOnReady = onready;
        _userOnDenied = ondenied;
        _userApplyTo = applyTo;

        var queryCriteria = new DevicePass.QueryCriteria();
        if(session)
            queryCriteria.session = session;
        queryCriteria.mode = mode;
        queryCriteria.lockToDomain = lockToDomain;
        queryCriteria.includeDeviceInfo = (mode== DevicePass.MODE_REGISTER_WITH_INFO);

        PBRPC.post('query_device_print', queryCriteria, function(response){
                if(response.error.code==200) {
                    _OnReady(response.serialized_response);
                }else if(response.error.code==401){
                    var rejection = DevicePass.Rejection.decode(response.serialized_response);

                    if (!window.location.origin) {
                        window.location.origin = window.location.protocol + "//" + window.location.hostname + (window.location.port ? ':' + window.location.port: '');
                    }
                    var url = "https://" + DevicePass.fqdn + ":" + DevicePass.port + "/prompt?domain="+ window.location.origin + "&cid=" + rejection.cid;
                    if(onpopup)
                        onpopup.call(_userApplyTo, url);
                }else if(response.error.code == 403){
                    _OnDenied();
                }
            },
            function(e){
                onerror.call(_userApplyTo, e);
            });
    };

    return DevicePass;

});