(function(global, factory) {

    /* AMD */ if (typeof define === 'function' && define["amd"])
        define(["protobuf"], factory);
    /* CommonJS */ else if (typeof require === "function" && typeof module === "object" && module && module["exports"])
        module["exports"] = factory(require("protobuf"), true);
    /* Global */ else
        (global["deepnet"] = global["deepnet"] || {})["PBRPC"] = factory(global["dcodeIO"]["ProtoBuf"]);

})(this, function(ProtoBuf) {

    "use strict";

    var PBRPC = {};

    PBRPC.VERSION = "1.0";

    PBRPC.proto_root = '/'; //require.toUrl('protos/');

    //PBRPC.builder = ProtoBuf.newBuilder();

    PBRPC.initialize = function(resourceBase, onInit){

        PBRPC.proto_root = resourceBase + "/lib/protos/";

        PBRPC.loadFile('rpc',
            function(err, bldr){
                if(err) {
                    console.log("err=" + err);
                    return;
                }

                PBRPC.Rpc         = bldr.build('device_manager.proto.Rpc');
                PBRPC.Request     = bldr.build('device_manager.proto.Request');
                PBRPC.Response    = bldr.build('device_manager.proto.Response');
                PBRPC.Error       = bldr.build('device_manager.proto.Error');
                if(onInit)
                    onInit();
            }
        );
    };

    PBRPC.loadFile = function(apiName, callback) {
        ProtoBuf.loadProtoFile(PBRPC.proto_root + apiName + '.proto',
            callback);
    };


    var calls = [];
    var events = [];

    function handleEvent(event){
        events.forEach(function(it){
           if(it.name===event.name)
               it.onEvent(event.serialized_parameters);
        });
    }

    function findAndRemoveCall(id){
        for (var i = 0, len = calls.length; i < len; i++){
            if(calls[i].id===id){
                var c = calls[i];
                calls.splice(i);
                return c;
            }
        }
        return null;
    }

    PBRPC.connect = function(url, onopen, onerror) {
        var ws = new WebSocket(url);
        PBRPC.ws = ws;

        if (!window.console) {
            window.console = {
                log: function () {
                }
            }
        }

        ws.onopen = function (ev) {
            console.log("ws open: " + ev);
            onopen();
        };
        ws.onerror = function (ev) {
            console.log("ws error: " + ev);
            var c = findAndRemoveCall(rpc.response.id);
            if(c){
                c.onError(ev); // call error
            }else{
                onerror(ev); // connect error
            }
        };
        ws.onclose = function (ev) {
            console.log("ws close: " +ev);
        };

        function handleResponse(response){

            // Decode the Message
            var rpc = PBRPC.Rpc.decode(response);

            if(rpc.response) {

                var c = findAndRemoveCall(rpc.response.id);
                if (c) {
                    c.onOK(rpc.response);
                }
            }else if(rpc.event){
                handleEvent(rpc.event);
            }

        }

        ws.onmessage = function (ev) {

            if(typeof ev.data == "string"){
                handleResponse(ev.data);
            }else {
                // conver blob to ArrayBuffer
                var fileReader = new FileReader();
                fileReader.onload = function () {
                    handleResponse(this.result);
                };
                fileReader.readAsArrayBuffer(ev.data);
            }
        }
    };

    var callerId = 1;

    PBRPC.post = function (method, params, onOK, onError) {
        var rpc = new PBRPC.Rpc();
        rpc.request = new PBRPC.Request();
        rpc.request.id = callerId;
        rpc.request.version = PBRPC.VERSION;
        rpc.request.method = method;
        if(params)
            rpc.request.serialized_request = params.toArrayBuffer();
        var buffer = rpc.toArrayBuffer();

        calls.push(
            {
                id:callerId,
                method:method,
                onOK:onOK,
                onError:onError
            }
        );

        callerId++;

        PBRPC.ws.send(buffer);
    };

    PBRPC.registerEventListener = function(name, onEvent){
        events.push({
            name:name,
            onEvent:onEvent
        });
    };

    return PBRPC;

});
